"""
This is the core of what will at some point become the new postprocessing
system in CosmoSIS.  The plan is:
- do plotting with GetDist instead of manually
- make the plotting much simpler than the complex hierarchies we have now
- connect this to the campaign system to make it easy to specify what to do after completing a campaign

These APIs will change and should not yet be relied on.

"""
from cosmosis.runtime.utils import import_by_path
from cosmosis import Inifile, output as output_module
import numpy as np
from io import StringIO
import os


class Chain:
    chain_subclasses = {}

    def __init__(
        self,
        name: str,
        sampler: str,
        columns: list[str],
        data: list[np.ndarray],
        metadata: list[dict[str, str]],
        comments: list[list[str]],
        final_metadata: list[list[str]],
        **options
    ):
        self.name = name
        self.sampler = sampler
        self.colnames = columns
        self.data = data
        self.metadata = metadata
        self.comments = comments
        self.final_metadata = final_metadata
        self.options = options
        self._mcsamples = None

    @classmethod
    def load(cls, inputs, name=None, **options):
        if isinstance(inputs, Inifile):
            name_, sampler, colnames, data, metadata, comments, final_metadata = (
                cls.load_ini(inputs)
            )
        elif isinstance(inputs, str):
            name_, sampler, colnames, data, metadata, comments, final_metadata = (
                cls.load_text_file(inputs)
            )
        elif isinstance(inputs, output_module.InMemoryOutput):
            name_, sampler, colnames, data, metadata, comments, final_metadata = (
                cls.load_in_memory_storage(inputs)
            )
        elif isinstance(inputs, output_module.AstropyOutput):
            name_, sampler, colnames, data, metadata, comments, final_metadata = (
                cls.load_astropy(inputs)
            )
        elif isinstance(inputs, output_module.OutputBase):
            name_, sampler, colnames, data, metadata, comments, final_metadata = (
                cls.load_output_object(inputs)
            )
        else:
            raise ValueError("Unknown input chain type " + str(type(inputs)))

        subclass = cls.chain_subclasses.get(sampler, cls)

        if name is None:
            name = name_

        # Assume metropolis if no more info is given
        if subclass is None:
            subclass = MetropolisChain

        return subclass(
            name, sampler, colnames, data, metadata, comments, final_metadata, **options
        )

    @classmethod
    def load_ini(cls, inputs):
        output_options = dict(inputs.items("output"))
        filename = output_options["filename"]
        name = filename
        sampler = inputs.get("runtime", "sampler")
        colnames, data, metadata, comments, final_metadata = (
            output_module.input_from_options(output_options)
        )
        return name, sampler, colnames, data, metadata, comments, final_metadata

    @classmethod
    def load_in_memory_storage(cls, inputs):
        if name is None:
            name = "chain"
        colnames = [c[0] for c in inputs.columns]
        data = [np.array(inputs.rows)]
        metadata = [{k: v[0] for k, v in inputs.meta.items()}]
        name = metadata[0].get("chain_name", "chain")
        sampler = metadata[0].get("sampler")
        final_metadata = [{k: v[0] for k, v in inputs.final_meta.items()}]
        comments = [inputs.comments[:]]
        return name, sampler, colnames, data, metadata, comments, final_metadata

    @classmethod
    def load_astropy(cls, inputs):
        name = inputs.meta.get("chain_name", "chain")
        colnames = inputs.colnames
        # convert astropy table to numpy array
        data = [np.array([inputs[c] for c in colnames]).T]
        metadata = [inputs.meta]
        sampler = metadata.get("sampler")
        final_metadata = [
            {k[6:]: v for k, v in meta.items() if k.startswith("final:")}
            for meta in metadata
        ]
        comments = [meta["comments"] for meta in metadata]
        return name, sampler, colnames, data, metadata, comments, final_metadata

    @classmethod
    def load_text_file(cls, inputs):
        colnames, data, metadata, comments, final_metadata = (
            output_module.TextColumnOutput.load_from_options({"filename": inputs})
        )
        name = metadata[0].get("chain_name", "chain")
        sampler = metadata[0].get("sampler")
        return name, sampler, colnames, data, metadata, comments, final_metadata

    @classmethod
    def load_output_object(self, inputs):
        colnames, data, metadata, comments, final_metadata = inputs.load_from_options(
            {"filename": inputs}
        )
        name = metadata[0].get("chain_name", "chain")
        sampler = metadata[0].get("sampler")
        return name, sampler, colnames, data, metadata, comments, final_metadata

    def __init_subclass__(cls):
        name = cls.__name__
        sampler = name.lower().replace("chain", "")
        cls.chain_subclasses[sampler] = cls

    def derive_extra_column(self, function):
        new_data = []
        for d in self.data:
            chain = SingleChainData(d, self.colnames)
            col, code = function(chain)
            if col is None:
                break
            # insert a new column into the chain, second from the end
            d = np.insert(d, -2, col, axis=1)
            # save the new chain
            new_data.append(d)
        if code is None:
            return

        self.colnames.insert(-2, code)
        self.data = new_data

    def derive_extra_columns(self):
        if not self.derive_file:
            return
        name = os.path.splitext(os.path.split(self.derive_file)[1])[0]
        module = import_by_path(name, self.derive_file)
        functions = [getattr(module, f) for f in dir(module) if f.startswith("derive_")]
        print(
            "Deriving new columns from these functions in {}:".format(self.derive_file)
        )
        for f in functions:
            self.derive_extra_column(f)

        # derive any additional parameters
        self.derive_extra_columns()

        # set the column names
        self.colnames = [c.lower() for c in self.colnames]
        self.data_stacked = np.concatenate(self.data).T

    def extract_ini(self, tag):
        in_ = False
        lines = []
        for line in self.comments[0]:
            line = line.strip()
            if line == "START_OF_{}_INI".format(tag):
                in_ = True
            elif line == "END_OF_{}_INI".format(tag):
                break
            elif in_:
                lines.append(line)

        s = StringIO("\n".join(lines))
        ini = Inifile(None)
        ini.read_file(s)
        return ini

    def __len__(self):
        return self.data_stacked.shape[1]

    def get_row(self, index):
        return self.data_stacked[:, index]

    def has_col(self, name):
        return name in self.colnames

    def get_col(self, index_or_name, stacked=True):
        """Get the named or numbered column."""
        if isinstance(index_or_name, int):
            index = index_or_name
        else:
            name = index_or_name
            index = self.colnames.index(name)
        cols = [d[:, index] for d in self.data]
        if stacked:
            return np.concatenate(cols)
        else:
            return cols

    @property
    def mcsamples(self):
        import getdist

        if self._mcsamples is None:
            samples = []
            names = []
            for col in self.colnames:
                if col == "weight":
                    continue
                c = self.reduced_col(col)
                samples.append(c)
                names.append(col)
            if "weight" in self.colnames:
                weights = self.reduced_col("weight")
            else:
                weights = None
            samples = np.array(samples).T
            self._mcsamples = getdist.MCSamples(
                samples=samples, weights=weights, names=names,
                name_tag=self.name
            )
        return self._mcsamples

    # Subclasses implement these
    def reduced_col(self, column):
        pass


class EmceeChain(Chain):
    def reduced_col(self, name, stacked=True):
        cols = self.get_col(name, stacked=False)
        burn = self.options.get("burn", 0)
        thin = self.options.get("thin", 1)

        if 0.0 < burn < 1.0:
            burn = int(len(cols[0]) * burn)
        else:
            burn = int(burn)

        cols = [col[burn::] for col in cols]

        if thin != 1:
            walkers = self.sampler_option("walkers")
            index = np.arange(len(cols[0]), dtype=np.int64)
            index = index // int(walkers)
            w = (index % thin) == 0
            cols = [col[w] for col in cols]

        if stacked:
            return np.concatenate(cols)
        else:
            return cols


class MetropolisChain(Chain):
    def reduced_col(self, name, stacked=True):
        cols = self.get_col(name, stacked=False)
        burn = self.options.get("burn", 0)
        thin = self.options.get("thin", 1)

        if 0.0 < burn < 1.0:
            burn = int(len(cols[0]) * burn)
        else:
            burn = int(burn)

        cols = [col[burn::thin] for col in cols]

        if stacked:
            return np.concatenate(cols)
        else:
            return np.array(cols).squeeze()


class ImportanceChain(MetropolisChain):
    # Importance sample chains may or may not
    # be based on a Metropolis-Hastings chain,
    # but they are treated the same way here.
    pass


class NestedChain(Chain):
    def reduced_col(self, name, stacked=True):
        """
        Nested sampling does not required cutting the chain
        from the main output file.

        These are also single chains, unlike the MH ones,
        so there is no stacking needed.
        """
        # stacking does not
        col = self.get_col(name)
        return np.array(col)


class PolychordChain(NestedChain):
    pass


class MultinestChain(NestedChain):
    pass


class ZeusChain(EmceeChain):
    pass


class PocoChain(MetropolisChain):
    pass


class DynestyChain(NestedChain):
    pass


class NautilusChain(NestedChain):
    pass


class SimpleListChain(Chain):
    """
    Base class for chains that are just lists of samples
    in some way. We might still want to burn or thin them though
    """

    def reduced_col(self, name, stacked=True):
        cols = self.get_col(name, stacked=False)
        thin = self.options.get("thin", 1)
        burn = self.options.get("burn", 0)
        if 0.0 < burn < 1.0:
            burn = int(len(cols[0]) * burn)
        else:
            burn = int(burn)
        cols = [col[burn::thin] for col in cols]
        return np.array(cols).squeeze()


class ListChain(SimpleListChain):
    pass


class AprioriChain(SimpleListChain):
    pass


class SingleChainData(object):
    """
    This helper object is to make it easier for users to write functions
    that derive new parameters.
    """

    def __init__(self, data, colnames):
        self.data = data
        self.colnames = colnames

    def __getitem__(self, index_or_name):
        if isinstance(index_or_name, int):
            index = index_or_name
        else:
            name = index_or_name
            index = self.colnames.index(name)
        return self.data[:, index]



import ctypes as ct

DBT_INT       =  0
DBT_DOUBLE    =  1
DBT_COMPLEX   =  2
DBT_STRING    =  3
DBT_INT1D     =  4
DBT_DOUBLE1D  =  5
DBT_COMPLEX1D =  6
DBT_STRING1D  =  7
DBT_BOOL      =  8
DBT_INTND     =  9
DBT_DOUBLEND  = 10
DBT_COMPELXND = 11
DBT_UNKNOWN   = 12



from cosmosis import Inifile, Sampler, LikelihoodPipeline, InMemoryOutput, TextColumnOutput
from cosmosis.postprocessing import postprocessor_for_sampler
from cosmosis.runtime import logs
import tempfile
import os
import sys
import pytest
import numpy as np
from astropy.table import Table


minuit_compiled = os.path.exists(Sampler.get_sampler("minuit").libminuit_name)

# our test priors are uniform on [-3, 3] for two
# parameters, so our expected prior is 1/6 for each of them.
EXPECTED_LOG_PRIOR = 2*np.log(1./6)

def run(name, check_prior, check_extra=True, can_postprocess=True, do_truth=False, no_extra=False, pp_extra=True, pp_2d=True, hints_peak=True, hints_cov=True, **options):

    sampler_class = Sampler.registry[name]

    values = tempfile.NamedTemporaryFile('w')
    values.write(
        "[parameters]\n"
        "p1=-3.0  0.0  3.0\n"
        "p2=-3.0  0.0  3.0\n")
    values.flush()

    override = {
        ('runtime', 'root'): os.path.split(os.path.abspath(__file__))[0],
        ('runtime', 'verbosity'): "noisy",
        ("pipeline", "debug"): "F",
        ("pipeline", "modules"): "test1",
        ("pipeline", "extra_output"): "parameters/p3",
        ("pipeline", "values"): values.name,
        ("test1", "file"): "example_module.py",
    }

    for k,v in options.items():
        override[(name,k)] = str(v)

    if no_extra:
        del override[("pipeline", "extra_output")]


    ini = Inifile(None, override=override)

    # Make the pipeline itself
    pipeline = LikelihoodPipeline(ini)

    output = InMemoryOutput()
    sampler = sampler_class(ini, pipeline, output)
    sampler.config()


    while not sampler.is_converged():
        sampler.execute()

    if check_prior:
        pr = np.array(output['prior'])
        # a few samples might be outside the bounds and have zero prior
        assert np.all((pr==EXPECTED_LOG_PRIOR)|(pr==-np.inf))
        # but not all of them
        assert not np.all(pr==-np.inf)

    if hints_peak:
        assert sampler.distribution_hints.has_peak()
        peak = sampler.distribution_hints.get_peak()
        assert peak.shape == (2,)
        idx = output["post"].argmax()
        assert np.isclose(output["parameters--p1"][idx], peak[0])
        assert np.isclose(output["parameters--p2"][idx], peak[1])
        sampler.distribution_hints.del_peak()
        assert not sampler.distribution_hints.has_peak()
    if hints_cov:
        assert sampler.distribution_hints.has_cov()
        cov = sampler.distribution_hints.get_cov()
        assert cov.shape == (2, 2)



    if check_extra and not no_extra:
        p1 = output['parameters--p1']
        p2 = output['parameters--p2']
        p3 = output['PARAMETERS--P3']
        assert np.all((p1+p2==p3)|(np.isnan(p3)))

    if can_postprocess:
        pp_class = postprocessor_for_sampler(name)
        print(pp_class)
        for getdist in [True, False]:
            with tempfile.TemporaryDirectory() as dirname:
                truth_file = values.name if do_truth else None
                pp = pp_class(output, "Chain", 0, outdir=dirname, prefix=name, truth=truth_file, fatal_errors=True, getdist=getdist)
                pp_files = pp.run()
                pp.finalize()
                pp.save()
                for p in pp_files:
                    print(p)
                postprocess_files = ['parameters--p1', 'parameters--p2']
                if pp_2d:
                    postprocess_files.append('2D_parameters--p2_parameters--p1')
                if check_extra and pp_extra and not no_extra:
                    postprocess_files.append('parameters--p3')
                    if pp_2d:
                        postprocess_files += ['2D_parameters--p3_parameters--p2', '2D_parameters--p3_parameters--p1']
                for p in postprocess_files:
                    filename = f"{dirname}{os.path.sep}{name}_{p}.png"
                    print("WANT ", filename)
                    assert filename in pp_files
                    assert os.path.exists(filename)
                for p in os.listdir(dirname):
                    p = os.path.join(dirname, p)
                    if p.endswith(".txt"):
                        Table.read(p, format='ascii.commented_header')
                        print(f"Read file {p} as a table")


    return output


def test_apriori():
    run('apriori', True, can_postprocess=False, hints_cov=False, nsample=100)

def test_dynesty():
    # dynesty does not support extra params
    run('dynesty', False, check_extra=False, nlive=25, sample='unif')

def test_emcee():
    run('emcee', True, walkers=8, samples=25)
    run('emcee', True, walkers=8, samples=25, a=3.0)

def test_truth():
    run('emcee', True, walkers=8, samples=25, do_truth=True)

def test_fisher():
    run('fisher', False, check_extra=False, hints_peak=False)
    run('fisher', False, check_extra=False, hints_peak=False)

def test_fisher_numdifftools():
    try:
        import numdifftools
    except ImportError:
        pytest.skip("numdifftools not installed")
    run('fisher', False, check_extra=False, hints_peak=False, method="numdifftools")

def test_fisher_smoothing():
    try:
        import derivative
    except ImportError:
        pytest.skip("derivative not installed")
    run('fisher', False, check_extra=False, hints_peak=False, method="smoothing")

def test_grid():
    run('grid', True, pp_extra=False, nsample_dimension=10, hints_cov=False)

def test_gridmax():
    run('gridmax', True, can_postprocess=False, max_iterations=1000, hints_cov=False)

# def test_kombine():
#     run('kombine')

def test_maxlike_single():
    output = run('maxlike', True, can_postprocess=False, hints_cov=False)
    assert len(output["post"]) == 1

def test_maxlike_alt():
    # alternative sampler, max-post, output_cov
    with tempfile.TemporaryDirectory() as dirname:
        output_ini = os.path.join(dirname, "output.ini")
        output_cov = os.path.join(dirname, "output_cov.txt")
        output_block = os.path.join(dirname, "output_block")
        run('maxlike', True, can_postprocess=False, method="L-BFGS-B", max_posterior=True, output_ini=output_ini, output_covmat=output_cov, output_block=output_block)
        assert os.path.exists(output_cov)
        assert os.path.exists(output_ini)
        assert os.path.isdir(output_block)
        


def test_maxlike_start_prior_repeat(caplog):
    output = run('maxlike', True, can_postprocess=False, repeats=5, start_method="prior", hints_cov=False)
    assert len(output["post"]) == 5
    assert (np.diff(output["like"]) >= 0).all()
    assert "Starting at a random point in the prior" in caplog.text

def test_maxlike_start_unable_to_repeat():
    # error - no start method specified but need one for repeats
    with pytest.raises(ValueError):
        run('maxlike', True, can_postprocess=False, repeats=5)

def test_maxlike_reiterations():
    # re-run the maxlike starting from the end of the previous run
    output = run('maxlike', True, can_postprocess=False, hints_cov=False, reiterations=2)
    assert len(output["post"]) == 1


def test_maxlike_start_no_start_file():
    # error - no start_input specified
    with pytest.raises(ValueError):
        run('maxlike', True, can_postprocess=False, repeats=5, start_method="chain")

def test_maxlike_start_chain_sample_auto(caplog):
    # Check we can start from a chain file
    with tempfile.NamedTemporaryFile('w') as f:
        f.write("#p1 p2 weight post\n")
        f.write("0.0 0.1  1.0  0.0\n")
        f.write("0.05 0.0  2.0  1.0\n")
        f.write("-0.1 0.2  2.0  2.0\n")
        f.flush()

        # This should work, and default to chain-sample because we have set repeats > 1
        run('maxlike', True, can_postprocess=False, repeats=5, start_method="chain", start_input=f.name, hints_cov=False)
        assert "Starting at random sample of points from chain file" in caplog.text

def test_maxlike_start_chain_maxpost_auto(caplog):
    # Check we can start from a chain file
    with tempfile.NamedTemporaryFile('w') as f:
        f.write("#p1 p2 weight post\n")
        f.write("0.0 0.1  1.0  0.0\n")
        f.write("0.05 0.0  2.0  1.0\n")
        f.write("-0.1 0.2  2.0  2.0\n")
        f.flush()

        # This should work, and default to chain-sample because we have set repeats > 1
        run('maxlike', True, can_postprocess=False, repeats=1, start_method="chain", start_input=f.name, hints_cov=False)
        assert "Starting at best posterior point from chain file" in caplog.text

def test_maxlike_start_chain_maxlike_auto(caplog):
    # Check we can start from a chain file
    with tempfile.NamedTemporaryFile('w') as f:
        f.write("#p1 p2 weight like\n")
        f.write("0.0 0.1  1.0  0.0\n")
        f.write("0.05 0.0  2.0  1.0\n")
        f.write("-0.1 0.2  2.0  2.0\n")
        f.flush()

        # This should work, and default to chain-sample because we have set repeats > 1
        run('maxlike', True, can_postprocess=False, repeats=1, start_method="chain", start_input=f.name, hints_cov=False)
        assert "Starting at best likelihood point from chain file" in caplog.text


def test_maxlike_start_chain_sample_manual(caplog):
    # Check we can start from a chain file
    with tempfile.NamedTemporaryFile('w') as f:
        f.write("#p1 p2 weight post\n")
        f.write("0.0 0.1  1.0  0.0\n")
        f.write("0.05 0.0  2.0  1.0\n")
        f.write("-0.1 0.2  2.0  2.0\n")
        f.flush()

        # This should work, and default to chain-sample because we have set repeats > 1
        run('maxlike', True, can_postprocess=False, repeats=5, start_method="chain-sample", start_input=f.name, hints_cov=False)
        assert "Starting at random sample of points from chain file" in caplog.text

    # Again with log-weights
    with tempfile.NamedTemporaryFile('w') as f:
        f.write("#p1 p2 weight post\n")
        f.write("0.0 0.1  0.0  0.0\n")
        f.write("0.05 0.0  0.7  1.0\n")
        f.write("-0.1 0.2  0.7  2.0\n")
        f.flush()

        # This should work, and default to chain-sample because we have set repeats > 1
        run('maxlike', True, can_postprocess=False, repeats=5, start_method="chain-sample", start_input=f.name, hints_cov=False)
        assert "Starting at random sample of points from chain file" in caplog.text

    # Again with no weights
    with tempfile.NamedTemporaryFile('w') as f:
        f.write("#p1 p2 post\n")
        f.write("0.0 0.1  0.0\n")
        f.write("0.05 0.0  1.0\n")
        f.write("-0.1 0.2   2.0\n")
        f.flush()

        # This should work, and default to chain-sample because we have set repeats > 1
        run('maxlike', True, can_postprocess=False, repeats=5, start_method="chain-sample", start_input=f.name, hints_cov=False)
        assert "Starting at random sample of points from chain file" in caplog.text


def test_maxlike_start_chain_fail(caplog):
    # Check we can start from a chain file
    with tempfile.NamedTemporaryFile('w') as f:
        f.write("#p1 p2 weight post\n")
        f.write("0.0 0.1  1.0  0.0\n")
        f.write("0.05 0.0  2.0  1.0\n")
        f.write("-0.1 0.2  2.0  2.0\n")
        f.flush()

        # should fail - no like column
        with pytest.raises(ValueError):
            run('maxlike', True, can_postprocess=False, repeats=1, start_method="chain-maxlike", start_input=f.name, hints_cov=False)


def test_maxlike_start_max_post(caplog):
    # Check we can start from a chain file
    with tempfile.NamedTemporaryFile('w') as f:
        f.write("#p1 p2 weight post\n")
        f.write("0.0 0.1  1.0  0.0\n")
        f.write("0.05 0.0  2.0  1.0\n")
        f.write("-0.1 0.2  2.0  2.0\n")
        f.flush()

        # start from best element of chain - should work because there is a max-post column
        run('maxlike', True, can_postprocess=False, repeats=1, start_method="chain-maxpost", start_input=f.name, hints_cov=False)
        assert "Starting at best posterior point from chain file" in caplog.text

def test_maxlike_start_last(caplog):
    # Check we can start from a chain file
    with tempfile.NamedTemporaryFile('w') as f:
        f.write("#p1 p2 weight post\n")
        f.write("0.0 0.1  1.0  0.0\n")
        f.write("0.05 0.0  2.0  1.0\n")
        f.write("-0.1 0.2  2.0  2.0\n")
        f.flush()

        # start from best element of chain - should work because there is a max-post column
        run('maxlike', True, can_postprocess=False, repeats=1, start_method="chain-last", start_input=f.name, hints_cov=False)
        assert "Starting from last point in file" in caplog.text



def test_maxlike_start_covmat(caplog):
    # Check we can start from a covmat
    with tempfile.NamedTemporaryFile('w') as f:
        f.write("0.1  0.0\n")
        f.write("0.0 0.08\n")
        f.flush()
        run('maxlike', True, can_postprocess=False, repeats=5, start_method="cov", start_input=f.name, hints_cov=False)
        assert "Starting at a random sample of points from the covariance of chain" in caplog.text


def test_start_estimate():
    values = tempfile.NamedTemporaryFile('w')
    values.write(
        "[parameters]\n"
        "p1=-10.0  0.0  10.0\n"
        "p2=-10.0  0.0  10.0\n")
    values.flush()

    override = {
        # ('runtime', 'verbosity'): "noisy",
        ("pipeline", "modules"): "",
        ("pipeline", "values"): values.name,
    }

    # mock objects just to stop the init breaking
    ini = Inifile(None, override=override)

    pipeline = LikelihoodPipeline(ini)
    sampler = Sampler(ini, pipeline)

    with tempfile.NamedTemporaryFile('w') as f:
        f.write("#p1    p2  weight like post\n")
        f.write(" 1.0   1.0  1.0  1.0   0.0\n") # non-zero weight
        f.write(" 2.0   2.0  0.0  2.0   0.0\n") # max-like
        f.write(" 3.0   3.0  0.0  1.0   1.0\n")
        f.write(" 4.0   4.0  0.0  0.0   2.0\n") # max-post
        f.write(" 5.0   5.0  0.0  0.0   0.0\n") # last
        f.flush()

        # only one sample has any weight so this one should be selected
        p = sampler.start_estimate(method="chain-sample", input_source=f.name, prefer_random=False)
        assert np.allclose(p, [1.0, 1.0])

        # only one sample has any weight so this one should be selected
        p = sampler.start_estimate(method="chain-maxlike", input_source=f.name, prefer_random=False)
        assert np.allclose(p, [2.0, 2.0])

        # last sample should be selected
        p = sampler.start_estimate(method="chain-maxpost", input_source=f.name, prefer_random=False)
        assert np.allclose(p, [4.0, 4.0])

        # last sample should be selected
        p = sampler.start_estimate(method="chain-last", input_source=f.name, prefer_random=False)
        assert np.allclose(p, [5.0, 5.0])

        # fall back to ini file choice
        p = sampler.start_estimate(method="", prefer_random=False)
        assert np.allclose(p, [0.0, 0.0])





def test_bobyqa():
    with tempfile.TemporaryDirectory() as dirname:
        output_cov = os.path.join(dirname, "output_cov.txt")
        run('maxlike', True, can_postprocess=False, method='bobyqa', output_covmat=output_cov)
        assert os.path.exists(output_cov)
    

def test_metropolis():
    run('metropolis', True, samples=20)
    run('metropolis', True, samples=20, covmat_sample_start=True)

@pytest.mark.skipif(not minuit_compiled,reason="requires Minuit2")
def test_minuit():
    run('minuit', True, can_postprocess=False, hints_cov=False)

def test_multinest():
    run('multinest', True, max_iterations=10000, live_points=50, feedback=False)

def test_pmaxlike():
    run('pmaxlike', True, can_postprocess=False, hints_cov=False)

def test_pmc():
    old_settings = np.seterr(invalid='ignore', divide='ignore')
    run('pmc', True, iterations=3, hints_cov=False)
    np.seterr(**old_settings)  

def test_zeus():
    run('zeus', True, maxiter=100_000, walkers=10, samples=100, nsteps=50, verbose=True)
    run('zeus', True, maxiter=100_000, walkers=10, samples=100, nsteps=50, moves="differential:2.0  global", tolerance=0.1, patience=5000)

def test_polychord():
    with tempfile.TemporaryDirectory() as base_dir:
        run('polychord', True, live_points=20, feedback=0, base_dir=base_dir, polychord_outfile_root='pc')

def test_snake():
        run('snake', True, pp_extra=False, hints_cov=False)


# Skip in a specific combination which causes a crash I can't track down.
# I think it's a memory thing as it's only when when in the full test suite
# not when run standalone.
@pytest.mark.skipif(os.environ.get("SKIP_NAUTILUS", "0")=="1", reason="nautilus runs out of memory on github actions")
def test_nautilus():
    run('nautilus', True, n_live=200, enlarge_per_dim=1.05,
        split_threshold=95., n_networks=2, n_batch=50, verbose=True, f_live=0.02, n_shell=50)


def test_star():
        run('star', False, pp_extra=False, pp_2d=False, hints_cov=False)

def test_test():
    run('test', False, can_postprocess=False, hints_peak=False, hints_cov=False)

def test_list_sampler():
    # test that the burn and thin parameters work
    # make a mock output file with some samples
    with tempfile.TemporaryDirectory() as dirname:

        input_chain_filename = os.path.join(dirname, "input_chain.txt")
        input_chain = TextColumnOutput(input_chain_filename)
        input_chain.add_column('parameters--p1', float)
        input_chain.add_column('parameters--p2', float)
        nrow = 100
        data = np.random.normal(size=(nrow, 2)).clip(-2.99, 2.99)
        for i in range(nrow):
            input_chain.parameters(data[i, 0], data[i, 1])
        input_chain.close()

        burn = 0.2
        thin = 2

        data = data[20::thin]

        output = run("list", True, can_postprocess=False, filename=input_chain_filename, burn=burn, thin=thin, hints_cov=False)
        p1 = output['parameters--p1']
        p2 = output['parameters--p2']
        p3 = p1 + p2
        expected_like = -(p1**2 + p2**2)/2.
        assert p1.size == 40
        assert np.allclose(p1, data[:, 0])
        assert np.allclose(p2, data[:, 1])
        assert np.allclose(output['PARAMETERS--P3'], p3)
        assert np.allclose(output["post"] - output["prior"], expected_like)

def test_importance_sampler():

    for add_to_likelihood in [True, False]:
        with tempfile.TemporaryDirectory() as dirname:

            input_chain_filename = os.path.join(dirname, "input_chain.txt")
            input_chain = TextColumnOutput(input_chain_filename)
            input_chain.add_column('parameters--p1', float)
            input_chain.add_column('parameters--p2', float)
            input_chain.add_column('post', float)
            nrow = 100
            original_posteriors = -np.arange(nrow).astype(float)
            data = np.random.normal(size=(nrow, 2)).clip(-2.99, 2.99)
            for i in range(nrow):
                input_chain.parameters(data[i, 0], data[i, 1], original_posteriors[i])
            input_chain.close()
            os.system(f"cat {input_chain_filename}")

            output = run("importance", True, can_postprocess=False, input=input_chain_filename, add_to_likelihood=add_to_likelihood, hints_cov=False)
            p1 = output['parameters--p1']
            p2 = output['parameters--p2']
            p3 = p1 + p2
            expected_like = -(p1**2 + p2**2)/2.
            # priors are uniform with range -3 to 3 so normalization should be np.log(1/6)
            expected_prior = EXPECTED_LOG_PRIOR
            expected_post = expected_like + expected_prior
            if add_to_likelihood:
                expected_post += original_posteriors
            assert p1.size == 100
            assert np.allclose(p1, data[:, 0])
            assert np.allclose(p2, data[:, 1])
            assert np.allclose(output['PARAMETERS--P3'], p3)
            assert np.allclose(output["post"], expected_post)

        

@pytest.mark.skipif(sys.version_info < (3,7), reason="pocomc requires python3.6+")
def test_poco():
    try:
        import pocomc
    except ImportError:
        pytest.skip("pocomc not installed")
    run('pocomc', True, check_extra=False, n_effective=32, n_active=16,  n_total=32, n_evidence=32)

if __name__ == '__main__':
    import sys
    locals()[sys.argv[1]]()


from .inputs import read_input
from .main import run_cosmosis_postprocess
from .postprocess import (
    postprocessor_for_sampler,
    MetropolisHastingsProcessor,
    EmceeProcessor,
    ZeusProcessor,
    KombineProcessor,
    PymcProcessor,
    MetropolisProcessor,
    WeightedMetropolisProcessor,
    ImportanceProcessor,
    GridProcessor,
    TestProcessor,
    MultinestProcessor,
    PolyChordProcessor,
    PMCPostProcessor,
    SnakeProcessor,
    FisherProcessor,
    StarProcessor,
)
from .plots import (
    GridPlots1D,
    GridPlots2D,
    SnakePlots2D,
    MetropolisHastingsPlots,
    TestPlots,
    WeightedPlots,
    MultinestPlots,
    PolychordPlots,
    WeightedMetropolisPlots,
    MCMCColorScatterPlot,
    WeightedMCMCColorScatterPlot,
    MultinestColorScatterPlot,
    PolychordColorScatterPlot,
    CovarianceMatrixGaussians,
    CovarianceMatrixEllipse,
    StarPlots,
    Tweaks,
    # These are deprecated:
    MetropolisHastingsPlots1D,
    WeightedMetropolisPlots1D,
    MultinestPlots1D,
    PolychordPlots1D,
    MetropolisHastingsPlots2D,
    WeightedMetropolisPlots2D,
    MultinestPlots2D,
    PolychordPlots2D,

)
from .reruns import BestFitRerunner, ini_from_header



from cosmosis.runtime.config import Inifile
from cosmosis.samplers.sampler import Sampler
import cosmosis.samplers.minuit.minuit_sampler
from cosmosis.runtime.pipeline import LikelihoodPipeline
from cosmosis.output import InMemoryOutput, TextColumnOutput
from cosmosis.runtime.handler import activate_segfault_handling
import tempfile
import os
import sys
import pytest
import numpy as np


activate_segfault_handling()

minuit_compiled = os.path.exists(cosmosis.samplers.minuit.minuit_sampler.libname)

def run(sampler, check_prior, inmem_output=True, check_extra=True, fast_slow=False, **options):

    sampler_class = Sampler.registry[sampler]

    values = tempfile.NamedTemporaryFile('w')
    values.write(
        "[parameters]\n"
        "p1=-3.0  0.0  3.0\n"
        "p2=-3.0  0.0  3.0\n"
        "p4=-3.0  0.0  3.0\n")
    values.flush()

    override = {
        ('runtime', 'root'): os.path.split(os.path.abspath(__file__))[0],
        ("pipeline", "debug"): "F",
        ("pipeline", "modules"): "test1  test3 test4",
        ("pipeline", "extra_output"): "parameters/p3",
        ("pipeline", "values"): values.name,
        ("pipeline", "fast_slow"): str(fast_slow)[0],
        ("pipeline", "first_fast_module"): "test4",
        ("test1", "file"): "example_module.py",
        ("test3", "file"): "example_module3.py",
        ("test4", "file"): "example_module4.py",
    }

    for k,v in options.items():
        override[(sampler,k)] = str(v)


    ini = Inifile(None, override=override)

    # Make the pipeline itself
    pipeline = LikelihoodPipeline(ini)

    if pipeline.do_fast_slow:
        pipeline.setup_fast_subspaces()

    if inmem_output:
        output = InMemoryOutput()
    else:
        output = TextColumnOutput(options["base_dir"] + "/output.txt")

    sampler = sampler_class(ini, pipeline, output)
    sampler.config()
    print("output = ", inmem_output, sampler.output, output)


    while not sampler.is_converged():
        sampler.execute()


    if inmem_output:
        if check_prior:
            pr = np.array(output['prior'])
            # but not all of them
            assert not np.all(pr==-np.inf)

        if check_extra:
            p1 = output['parameters--p1']
            p2 = output['parameters--p2']
            p3 = output['PARAMETERS--P3']
            assert np.all((p1+p2==p3)|(np.isnan(p3)))
    else:
        print(os.listdir(options["base_dir"]))
        assert os.path.exists(options["base_dir"] + "/output.txt")
        assert os.path.exists(options["base_dir"] + "/output_boosted.txt")

    return output


def test_polychord_hang():
    # with tempfile.TemporaryDirectory() as base_dir:
    #     run('polychord', True, live_points=20, feedback=0, base_dir=base_dir, polychord_outfile_root='pc')

    # Test for issue # https://bitbucket.org/joezuntz/cosmosis/issues/382/samplers-will-hang-if-fast_fraction-is-non
    with tempfile.TemporaryDirectory() as base_dir:
        run('polychord', True, live_points=20, feedback=0, base_dir=base_dir, polychord_outfile_root='pc', fast_slow=True, fast_fraction=0.5)


def test_polychord_boosting():
    with tempfile.TemporaryDirectory() as base_dir:
        run('polychord', True, inmem_output=False, live_points=20, boost_posteriors=2.0, feedback=0, base_dir=base_dir, polychord_outfile_root='pc', fast_slow=True, fast_fraction=0.5)


if __name__ == '__main__':
    import sys
    locals()[sys.argv[1]]()


#!/usr/bin/env python
import sys
import argparse
import os
from .postprocessing import run_cosmosis_postprocess

parser = argparse.ArgumentParser(description="Post-process cosmosis output")
parser.add_argument("inifile", nargs="+")
mcmc=parser.add_argument_group(title="MCMC", description="Options for MCMC-type samplers")
mcmc.add_argument("--burn", default=0.0, type=float, help="Fraction or number of samples to burn at the start")
mcmc.add_argument("--thin", default=1, type=int, help="Keep every n'th sampler in MCMC")
mcmc.add_argument("--weights", action='store_true', help="Look for a weight column in a generic MCMC file")
mcmc.add_argument("-gd","--getdist", action='store_true', default=False, help="Use getdist for the statistics of MCMC chain. Not implemented for polychord and grid sampler yet.")

general=parser.add_argument_group(title="General", description="General options for controlling postprocessing")
general.add_argument("-o","--outdir", default=".", help="Output directory for all generated files")
general.add_argument("-p","--prefix", default="", help="Prefix for all generated files")
general.add_argument("--more-latex", default="", help="Load an additional latex file to the default")
general.add_argument("--no-latex", action='store_true', help="Do not use latex-style labels, just use the text")
general.add_argument("--blind-add", action='store_true', help="Blind results by adding adding a secret value to each parameter")
general.add_argument("--blind-mul", action='store_true', help="Blind results by scaling by a secret value for each parameter")
general.add_argument("--pdb", action='store_true', help="Run the debugger if any of the postprocessing stages fail")
general.add_argument("--fatal-errors", action='store_true', help="Errors are fatal; useful for testing only")

inputs=parser.add_argument_group(title="Inputs", description="Options controlling the inputs to this script")
inputs.add_argument("--text", action='store_true', help="Tell postprocess that its argument is a text file, regardless of its suffix")
inputs.add_argument("--derive", default="", help="Read a python script with functions in that derive new columns from existing ones")

plots=parser.add_argument_group(title="Plotting", description="Plotting options")
plots.add_argument("--legend", help="Add a legend to the plot with the specified titles, separated by | (the pipe symbol)")
plots.add_argument("--legend-loc", default='best', help="The location of the legend: best, UR, UL, LL, LR, R, CL, CR, LC, UC, C (use quotes for the ones with two words.)")
plots.add_argument("--swap", action='store_true', help="Swap the ordering of the parameters in (x,y)")
plots.add_argument("--only", type=str, dest='prefix_only', help="Only make 2D plots where both parameter names start with this")
plots.add_argument("--either", type=str, dest='prefix_either', help="Only make 2D plots where one of the parameter names starts with this.")
parser.add_argument("--exclude", nargs="+", dest='prefix_exclude', help='Specify one or more prefixes to exclude matching parameters from 2D plots')

plots.add_argument("--no-plots", action='store_true', help="Do not make any default plots")
plots.add_argument("--no-2d", action='store_true', help="Do not make any 2D plots")
plots.add_argument("--no-alpha", dest='alpha', action='store_false', help="No alpha effect - shaded contours will not be visible through other ones")
plots.add_argument("-f", "--file-type", default="png", help="Filename suffix for plots")
plots.add_argument("--no-smooth", dest='smooth', default=True, action='store_false', help="Do not smooth grid plot joint constraints")
plots.add_argument("--fix-edges", dest='fix_edges', default=True, action='store_true', help="Use an alternative KDE to fix 1D plot boundaries")
plots.add_argument("--no-fix-edges", dest='fix_edges', default=False, action='store_false', help="Switch off the edge fixing")
plots.add_argument("--n-kde", default=100, type=int, help="Number of KDE smoothing points per dimension to use for MCMC 2D curves. Reduce to speed up, but can make plots look worse.")
plots.add_argument("--factor-kde", default=2.0, type=float, help="Smoothing factor for MCMC plots.  More makes plots look better but can smooth out too much.")
plots.add_argument("--no-fill", dest='fill', default=True, action='store_false', help="Do not fill in 2D constraint plots with color")
plots.add_argument("--extra", dest='extra', default="", help="Load extra post-processing steps from this file.")
plots.add_argument("--tweaks", dest='tweaks', default="", help="Load plot tweaks from this file.")
plots.add_argument("--no-image", dest='image', default=True, action='store_false', help="Do not plot the image in  2D grids; just show the contours")
plots.add_argument("--run-max-post", default="", help="Run the test sampler on maximum-posterior sample and save to the named directory.")
plots.add_argument("--truth", default="", help="An ini file containing truth values to mark on plots")


def main(args):
	args = parser.parse_args(args)

	# Run the main postprocessing code
	processor = run_cosmosis_postprocess(args.inifile, **vars(args))

	# The postprocessor doesn't finalize the plots until it is saved
	# so that we can use it interactively too. So save here.
	if processor is not None:
		#Save all the image files and close the text files
		processor.save()

if __name__=="__main__":
	main(sys.argv[1:])



from cosmosis.runtime import Inifile, register_new_parameter, LikelihoodPipeline, Parameter, Module
from cosmosis.datablock import DataBlock
from cosmosis.samplers.sampler import Sampler
from cosmosis.runtime.prior import TruncatedGaussianPrior, DeltaFunctionPrior
from cosmosis.output.in_memory_output import InMemoryOutput
from cosmosis.main import run_cosmosis, parser, mpi_pool
import numpy as np
import os
import tempfile
import pstats
import pytest

root = os.path.split(os.path.abspath(__file__))[0]

def test_add_param():

    sampler_class = Sampler.registry["emcee"]

    values = tempfile.NamedTemporaryFile('w')
    values.write(
        "[parameters]\n"
        "p1=-3.0  0.0  3.0\n"
        "p2=-3.0  0.0  3.0\n")
    values.flush()

    override = {
        ('runtime', 'root'): root,
        ("pipeline", "debug"): "F",
        ("pipeline", "modules"): "test2",
        ("pipeline", "values"): values.name,
        ("test2", "file"): "example_module2.py",
        ("emcee", "walkers"): "8",
        ("emcee", "samples"): "10"
    }

    ini = Inifile(None, override=override)

    # Make the pipeline itself
    pipeline = LikelihoodPipeline(ini)

    # test the the new added parameter has worked
    assert len(pipeline.varied_params) == 3
    p = pipeline.varied_params[2]
    assert str(p) == "new_parameters--p3"
    assert isinstance(p, Parameter)
    assert isinstance(p.prior, TruncatedGaussianPrior)
    assert np.isclose(p.prior.mu, 0.1)
    assert np.isclose(p.prior.sigma, 0.2)

    assert len(pipeline.fixed_params) == 1
    p = pipeline.fixed_params[0]
    assert isinstance(p, Parameter)
    assert isinstance(p.prior, DeltaFunctionPrior)


    output = InMemoryOutput()
    sampler = sampler_class(ini, pipeline, output)
    sampler.config()


    # check that the output is working
    assert output.column_index_for_name("new_parameters--p3") == 2

    while not sampler.is_converged():
        sampler.execute()


    p1 = output['parameters--p1']
    p2 = output['parameters--p2']
    p3 = output['new_parameters--p3']
    assert p3.max() < 1.0
    assert p3.min() > -1.0

def test_missing_setup():
    # check the register_new_parameter feature when no
    # setup is currently happening
    module = Module("test2", root + "/example_module2.py")
    config = DataBlock()
    module.setup(config)

def test_unused_param_warning(capsys):
    # check that an appropriate warning is generated
    # when a parameter is unused
    module = Module("test", root + "/example_module.py")
    config = DataBlock()
    config['test', 'unused'] = "unused_parameter"
    module.setup(config)
    out, _ = capsys.readouterr()
    assert "**** WARNING: Parameter 'unused'" in out

def test_vector_extra_outputs():
    with tempfile.TemporaryDirectory() as dirname:
        values_file = f"{dirname}/values.ini"
        params_file = f"{dirname}/params.ini"
        output_file = f"{dirname}/output.txt"
        with open(values_file, "w") as values:
            values.write(
                "[parameters]\n"
                "p1=-3.0  0.0  3.0\n"
                "p2=-3.0  0.0  3.0\n")

        params = {
            ('runtime', 'root'): os.path.split(os.path.abspath(__file__))[0],
            ('runtime', 'sampler'):  "emcee",
            ("pipeline", "debug"): "T",
            ("pipeline", "modules"): "test1",
            ("pipeline", "extra_output"): "data_vector/test_theory#2",
            ("pipeline", "values"): values_file,
            ("test1", "file"): "example_module.py",
            ("output", "filename"): output_file,
            ("emcee", "walkers"): "8",
            ("emcee", "samples"): "10",
        }

        ini = Inifile(None, override=params)
        status = run_cosmosis(ini)

        with open(output_file) as f:
            header = f.readline()

        assert "data_vector--test_theory_0" in header.lower()
        assert "data_vector--test_theory_1" in header.lower()

        data = np.loadtxt(output_file)
        # two parameters, two extra saves, prior, and posterior
        assert data.shape[1] == 6


def test_profile(capsys):
    with tempfile.TemporaryDirectory() as dirname:
        values_file = f"{dirname}/values.ini"
        params_file = f"{dirname}/params.ini"
        output_file = f"{dirname}/output.txt"
        stats_file  = f'{dirname}/profile.dat'
        with open(values_file, "w") as values:
            values.write(
                "[parameters]\n"
                "p1=-3.0  0.0  3.0\n"
                "p2=-3.0  0.0  3.0\n")

        params = {
            ('runtime', 'root'): os.path.split(os.path.abspath(__file__))[0],
            ('runtime', 'sampler'):  "emcee",
            ("pipeline", "debug"): "T",
            ("pipeline", "modules"): "test1",
            ("pipeline", "values"): values_file,
            ("test1", "file"): "example_module.py",
            ("output", "filename"): output_file,
            ("emcee", "walkers"): "8",
            ("emcee", "samples"): "10",
        }

        ini = Inifile(None, override=params)
        status = run_cosmosis(ini, profile_cpu=stats_file)

        output = capsys.readouterr()
        assert "cumtime" in output.out

        stats = pstats.Stats(stats_file)
        stats.sort_stats("cumtime")
        stats.print_stats(10)



def test_script_skip():
    with tempfile.TemporaryDirectory() as dirname:
        values_file = f"{dirname}/values.ini"
        params_file = f"{dirname}/params.ini"
        output_file = f"{dirname}/output.txt"
        with open(values_file, "w") as values:
            values.write(
                "[parameters]\n"
                "p1=-3.0  0.0  3.0\n"
                "p2=-3.0  0.0  3.0\n")

        params = {
            ('runtime', 'root'): os.path.split(os.path.abspath(__file__))[0],
            ('runtime', 'pre_script'): "this_executable_does_not_exist",
            ('runtime', 'post_script'): "this_executable_does_not_exist",
            ('runtime', 'sampler'):  "test",
            ("pipeline", "debug"): "F",
            ("pipeline", "modules"): "test1",
            ("pipeline", "values"): values_file,
            ("test1", "file"): "example_module.py",
            ("output", "filename"): output_file,
        }

        ini = Inifile(None, override=params)

        with pytest.raises(RuntimeError):
            status = run_cosmosis(ini)

        # shopuld work this time
        try:
            os.environ["COSMOSIS_NO_SUBPROCESS"] = "1"
            status = run_cosmosis(ini)
        finally:
            del os.environ["COSMOSIS_NO_SUBPROCESS"]

def test_prior_override():
    with tempfile.TemporaryDirectory() as dirname:
        values_file = f"{dirname}/values.ini"
        output_file = f"{dirname}/output.txt"
        with open(values_file, "w") as values:
            values.write(
                "[parameters]\n"
                "p1=-3.0  0.0  3.0\n"
                "p2=-3.0  0.0  3.0\n")

        params = {
            ('runtime', 'root'): os.path.split(os.path.abspath(__file__))[0],
            ('runtime', 'sampler'):  "apriori",
            ('apriori', 'nsample'):  "1000",
            ("pipeline", "debug"): "F",
            ("pipeline", "modules"): "test1",
            ("pipeline", "values"): values_file,
            ("test1", "file"): "example_module.py",
            ("output", "filename"): output_file,
        }

        args = parser.parse_args(["not_a_real_file"])
        ini = Inifile(None, override=params)

        priors_vals = {
            ('parameters', 'p1'): 'uniform -1.0    1.0'
        }

        priors = Inifile(None, override=priors_vals)

        status = run_cosmosis(ini, priors=priors)
        chain = np.loadtxt(output_file).T
        p1 = chain[0]
        p2 = chain[1]
        print(p1.min(), p1.max())
        assert p1.max() < 1.0
        assert p1.min() > -1.0

def test_pipeline_from_function():
    priors = {
        "p0": "gaussian 1.0 0.5",
    }

    def log_like(p):
        r1 = np.sum(np.abs(p))
        return -0.5 * np.sum(p**2), {"r1": r1}

    param_ranges = [
        (-3.0,  0.0,  3.0),
        (-3.0,  0.0,  3.0),
        (-3.0,  0.0,  3.0),
        (-3.0,  0.0,  3.0),
    ]

    derived = ["r1"]

    pipeline = LikelihoodPipeline.from_likelihood_function(log_like, param_ranges, priors=priors, derived=derived, debug=True)
    r = pipeline.run_results([0,0,0,0])
    assert r.like == 0.0
    assert r.extra[0] == 0.0

    # version without priors or derived params extracted
    pipeline = LikelihoodPipeline.from_likelihood_function(log_like, param_ranges)
    r = pipeline.run_results([0,0,0,0])
    assert r.like == 0.0
    assert len(r.extra) == 0


def test_failure_log():
    def log_like(p):
        if p[0] < 0:
            print("Deliberately failing pipeline ...")
            raise ValueError("p0 must be positive")
        return -0.5 * np.sum(p**2)

    param_ranges = [
        (-3.0,  0.0,  3.0),
        (-3.0,  0.0,  3.0),
        (-3.0,  0.0,  3.0),
        (-3.0,  0.0,  3.0),
    ]

    with tempfile.TemporaryDirectory() as dirname:
        logname = f"{dirname}/failure.log"
        failure_log = mpi_pool.MPILogFile(logname)
    
        pipeline = LikelihoodPipeline.from_likelihood_function(log_like, param_ranges)
        pipeline.set_failure_log_file(failure_log)
        # two that should work and produce nothing
        # and two that should log some errors
        param_sets = [
            [0.0,0.0,0.0,0.0],
            [1.0,0.0,0.0,0.0],
            [-1.0,0.0,0.0,4.0],
            [-2.0,0.0,0.0,8.0],
        ]
        print("Pipeline errors are expected below - we are testing logging of them")
        for p in param_sets:
            pipeline.run_parameters(p)

        failure_log.close()
        with open(logname) as f:
            lines = f.readlines()
        assert len(lines) == 2
        assert lines[0].strip() == "-1.0 0.0 0.0 4.0"
        assert lines[1].strip() == "-2.0 0.0 0.0 8.0"


def test_recreate_pipeline():
    with tempfile.TemporaryDirectory() as dirname:

        values = os.path.join(dirname, "values.ini")
        priors = os.path.join(dirname, "priors.ini")
        output = os.path.join(dirname, "output.txt")
        with open(values, "w") as f:
            f.write(
                "[parameters]\n"
                "p1=-10.0  0.0  10.0\n"
                "p2=-1000.0  0.0  1000.0\n")

        with open(priors, "w") as f:
            f.write(
                "[parameters]\n"
                "p1=uniform -5.0 5.0\n"
                "p2=gaussian 0.0 1.0"
            )

        override = {
            ('runtime', 'root'): root,
            ('runtime', 'sampler'): "emcee",
            ("pipeline", "debug"): "F",
            ("pipeline", "modules"): "test2",
            ("pipeline", "values"): values,
            ("pipeline", "priors"): priors,
            ("pipeline", "extra_output"): "parameters/p3",
            ("output", "filename"): output,
            ("output", "format"): "text",
            ("test2", "file"): "example_module.py",
            ("emcee", "walkers"): "8",
            ("emcee", "samples"): "10"
        }

        ini = Inifile(None, override=override)

        status = run_cosmosis(ini)

        # Now we want to recreate the pipeline from the output
        pipeline = LikelihoodPipeline.from_chain_file(output)

        # check the basic pipeline configuration
        assert pipeline.modules[0].name == "test2"
        assert len(pipeline.modules) == 1

        # check it produces the same results
        r = pipeline.run_results([1.,2.])
        assert np.isclose(r.like, -2.5)
        assert np.isclose(r.extra[0], 3.0)

        # check that the priors have been correctly passed through
        # the recreated pipeline
        assert np.isclose(r.prior, np.log(0.1) - 2.0 - 0.5*np.log(2*np.pi))




if __name__ == '__main__':
    test_script_skip()



from .. import ParallelSampler
from ...runtime import logs
import numpy as np
import scipy.optimize
import warnings

# The likelihood function we wish to optimize.
# Not that we are minimizing the negative of the likelihood/posterior
def likefn(p_in):
    global sampler
    # Check the normalization
    if (not np.all(p_in>=0)) or (not np.all(p_in<=1)):
        return np.inf
    p = sampler.pipeline.denormalize_vector(p_in)
    r = sampler.pipeline.run_results(p)
    if sampler.max_posterior:
        return -r.post
    else:
        return -r.like

def run_optimizer(start_vector):
    global sampler
    return sampler.run_optimizer(start_vector)

class MaxlikeSampler(ParallelSampler):
    parallel_output = False
    supports_resume = False # TODO: support resuming

    sampler_outputs = [("prior", float), ("like", float), ("post", float)]

    def config(self):
        global sampler
        sampler = self
        self.tolerance = self.read_ini("tolerance", float, 1e-3)
        self.maxiter = self.read_ini("maxiter", int, 1000)
        self.output_ini = self.read_ini("output_ini", str, "")
        self.output_cov = self.read_ini("output_covmat", str, "")
        self.output_block = self.read_ini("output_block", str, "")
        self.method = self.read_ini("method",str,"Nelder-Mead")
        self.max_posterior = self.read_ini("max_posterior", bool, False)
        self.repeats = self.read_ini("repeats", int, 1)
        self.reiterations = self.read_ini("reiterations", int, 1)
        if self.repeats == 1:
            nsteps = 1
        elif self.pool is not None:
            nsteps = self.pool.size
        else:
            nsteps = 1
        self.nsteps = self.read_ini("nsteps", int, nsteps)
        start_method = self.read_ini("start_method", str, "")

        if self.repeats > 1:
            start_method_is_random = start_method in ["prior", "cov", "chain", "chain-sample"]
            have_previous_peak = self.distribution_hints.has_peak()
            have_previous_cov = self.distribution_hints.has_cov()

            if have_previous_peak and not have_previous_cov:
                raise ValueError("You have set repeats>1 in maxlike, and have chained multiple samplers together. "
                                 "But the previous sampler(s) did not save a covariance matrix. "
                                 "So we have no way to generate multiple starting points. "
                                 )
            elif not start_method_is_random:
                raise ValueError(f"You set the repeats parameter in the max-like sampler to {self.repeats}"
                                 " but to you this you must also set a start method to choose a new starting point"
                                 " each repeat.  You can set start_method to 'prior' to sample from the prior,"
                                 " or set it to 'cov' or 'chain' and also set start_input to a file to load one"
                                 " either a covariance matrix or a chain of samples."
                                 )
        if self.is_master():
            if self.max_posterior:
                logs.overview("------------------------------------------------")
                logs.overview("NOTE: Running optimizer in **max-posterior** mode:")
                logs.overview("NOTE: Will maximize the combined likelihood and prior")
                logs.overview("------------------------------------------------")
            else:
                logs.overview("--------------------------------------------------")
                logs.overview("NOTE: Running optimizer in **max-like** mode:")
                logs.overview("NOTE: not including the prior, just the likelihood.")
                logs.overview("NOTE: Set the parameter max_posterior=T to change this.")
                logs.overview("NOTE: This won't matter unless you set some non-flat")
                logs.overview("NOTE: priors in a separate priors file.")
                logs.overview("--------------------------------------------------")

        self.best_fit_results = []
        self.start_points = None

    def save_final_outputs(self, best_fit_results, final=False):

        # This can happen if the user ctrl'c's the run before any results are saved
        if not best_fit_results:
            return

        # Sort the repeated best-fit estimates by increasing posterior or likelihood
        if self.max_posterior:
            best_fit_results.sort(key=lambda x: x.post)
        else:
            best_fit_results.sort(key=lambda x: x.like)

        # Save all the results to the main chain output file.
        # We will overwrite previous sets of results in the file here
        self.output.reset_to_chain_start()
        for results in best_fit_results:
            self.output.parameters(results.vector, results.extra, results.prior, results.like, results.post)

        # Get the overall best-fit results
        results = best_fit_results[-1]

        # Also if requested, approximate the covariance matrix with the
        # inverse of the Hessian matrix.
        # For a gaussian posterior this is exact.
        if results.covmat is None:
            if self.output_cov:
               warnings.warn("Sorry - the optimization method you chose does not return a covariance (or Hessian) matrix")
        else:
            if self.output_cov:
                np.savetxt(self.output_cov, results.covmat)

        if self.output_ini:
            self.pipeline.create_ini(results.vector, self.output_ini)
        
        if self.output_block:
            results.block.save_to_directory(self.output_block)


        # We only want to update the distribution hints at the very end
        if final:
            # These values are used by subsequent samplers, if you chain
            # some of them together
            self.distribution_hints.set_peak(results.vector, results.post)
            if results.covmat is not None:
                self.distribution_hints.set_cov(results.covmat)


    def run_optimizer(self, inputs):
        rank = self.pool.rank if self.pool is not None else 0
        start_vector, repeat_index = inputs
        start_vector_denorm = self.pipeline.denormalize_vector(start_vector)
        logs.overview(f"[Rank {rank}] Starting from point: {start_vector_denorm}")
        bounds = [(0.0, 1.0) for p in self.pipeline.varied_params]

        if self.method.lower() == "bobyqa":
            # use the specific pybobyqa minimizer
            import pybobyqa

            # this works with bounds in the form of a tuple of two arrays
            lower = np.array([b[0] for b in bounds])
            upper = np.array([b[1] for b in bounds])
            kw = {
                "seek_global_minimum": True,
                "bounds": (lower,upper),
                "print_progress": logs.is_enabled_for(logs.NOISY),
                "rhobeg": 0.1,
                "rhoend": self.tolerance,
            }
            for i in range(self.reiterations):
                optimizer_result = pybobyqa.solve(likefn, start_vector, **kw)
                start_vector = optimizer_result.x
            opt_norm = optimizer_result.x
            # bobyqa calls it .hessian but scipy calls it .hess, so copy it here
            # if available
            if optimizer_result.hessian is not None:
                optimizer_result.hess = optimizer_result.hessian
        else:
            # Use scipy mainimizer instead
            for i in range(self.reiterations):
                optimizer_result = scipy.optimize.minimize(likefn, start_vector, method=self.method,
                    jac=False, tol=self.tolerance,
                    options={'maxiter':self.maxiter, 'disp':True})
                start_vector = optimizer_result.x

            opt_norm = optimizer_result.x

        opt = self.pipeline.denormalize_vector(opt_norm)
        
        # Generate a results object containing the best-fit parameters,
        # their likelihood, posterior, prior, and derived parameters
        results = self.pipeline.run_results(opt)

        # Log some info to the screen
        par_text = '   '.join(str(x) for x in opt)
        logs.overview(f"\n\n[Rank {rank}] Optimization run {repeat_index+1} of {self.repeats}")
        if self.max_posterior:
            logs.overview(f"[Rank {rank}] Best fit (by posterior): {par_text}\n%s")
        else:
            logs.overview(f"[Rank {rank}] Best fit (by likelihood):\n{par_text}")
        logs.overview(f"[Rank {rank}] Posterior: {results.post}")
        logs.overview(f"[Rank {rank}] Likelihood: {results.like}\n")

        # Also attach any covariance matrix saved by the sampler to the results
        # object, in an ad-hoc way
        if hasattr(optimizer_result, 'hess_inv'):
            if self.method == "L-BFGS-B":
                results.covmat = self.pipeline.denormalize_matrix(optimizer_result.hess_inv.todense())
            else:
                results.covmat = self.pipeline.denormalize_matrix(optimizer_result.hess_inv)
        elif hasattr(optimizer_result, 'hess'):
            results.covmat = self.pipeline.denormalize_matrix(np.linalg.inv(optimizer_result.hess))
        else:
            results.covmat = None

        return results
    
    def choose_start(self, n):
        if n == 1:
            # we just want a single starting point. 
            # So we can use the basic one from the values file or
            # previous sampler; we don't have to worry about making
            # it randomized.

            # It's also possible that this is just the last of our repeats
            # to be done, but in that case it's also okay to just use the
            # basic starting point, since we won't already have used it
            start_vector_denorm = self.start_estimate()
            start_vector = self.pipeline.normalize_vector(start_vector_denorm)

            # If that is invalid we will raise an error
            if not np.isfinite(likefn(start_vector)):
                raise RuntimeError("Invalid starting point for maxlike")
            
            start_vectors = np.array([start_vector]) # 1 x n array

        else:
            start_vectors = np.zeros((n, self.pipeline.nvaried))
            for i in range(n):
                # If we are taking a random starting point then there is a chance it will randomly
                # be invalid. So we should try a 20 times to get a valid one.
                for j in range(20):
                    start_vector_denorm = self.start_estimate(prefer_random=True, quiet=(i>0 or j>0))
                    start_vector = self.pipeline.normalize_vector(start_vector_denorm)
                    if np.isfinite(likefn(start_vector)):
                        break
                else:
                    raise RuntimeError("Could not find a valid random starting point for maxlike in 20 tries")
                start_vectors[i] = start_vector

        return start_vectors

    def execute(self):
        # Choose a starting point. If we are repeating our optimization we will need
        # multiple starting points. Otherwise we use a fixed one.
        # If we are re-starting then this will actually generate a different set of
        # starting points compared to the first time, but they are random so it doesn't
        # matter
        if self.start_points is None:
            self.start_points = self.choose_start(self.repeats)

        # Figure out how many steps we need to do now.
        ndone = len(self.best_fit_results)
        if ndone + self.nsteps > self.repeats:
            n = self.repeats - ndone
        else:
            n = self.nsteps

        # Starting points to use for this chunk of steps
        starting_points = self.start_points[ndone:ndone + n]

        if self.pool is None:
            # serial mode. Just do everything on this process.
            # this allows us to also do a nice thing where if we
            # get part way through the results and then the user
            # ctrl-c's then we can still output partial results
            collected_results = []
            for i, start_vector in enumerate(starting_points):
                try:
                    results = self.run_optimizer((start_vector, ndone + i))
                    collected_results.append(results)
                except KeyboardInterrupt:
                    # If we get a ctrl-c we should save the current best fit
                    # and then exit.
                    # Otherwise report what we are doing. It should be basically instant
                    # so it shouldn't annoy anyone
                    logs.overview("Keyboard interrupt - saving current best fit, if any finished")
                    self.save_final_outputs(collected_results, final=True)
                    raise
        else:
            inputs = [(start_vector, ndone + i) for i, start_vector in enumerate(starting_points)]
            collected_results = self.pool.map(run_optimizer, inputs)

        # Add this to our list of best-fits
        self.best_fit_results.extend(collected_results)

        # we re-save the final outputs each time, rewinding the file
        # to overwrite the previous ones, so we can re-sort each time.
        self.save_final_outputs(self.best_fit_results, final=self.is_converged()    )

    def is_converged(self):
        return len(self.best_fit_results) >= self.repeats



from .output_base import OutputBase

class AstropyOutput(OutputBase):
    _aliases = ["astropy"]
    def __init__(self):
        super(AstropyOutput,self).__init__()
        from astropy.table import Table
        self.table = Table()

    def _begun_sampling(self, params):
        from astropy.table import Column
        for name, dtype, comment in self.columns:
            self.table.add_column(Column(name=name, dtype=dtype))

    def _write_parameters(self, params):
        self.table.add_row(params)

    def __getitem__(self, key_or_index):
        return self.table[key_or_index]

    def _write_metadata(self, key, value, comment):
            self.table.meta[key] = value
            if comment:
                self.table.meta[key + "_comment"] = comment

    def _write_final(self, key, value, comment):
            self.table.meta["final:" + key] = value
            if comment:
                self.table.meta["final:" + key + "_comment"] = comment

    def _write_comment(self, comment):
        self.table.meta["comments"] = self.table.meta.get("comments", []) + [comment]

    def reset_to_chain_start(self):
         self.table.remove_rows(slice(None))




import abc
import numpy as np
import fcntl

output_registry = {}

class OutputMetaclass(abc.ABCMeta):
    def __init__(cls, name, b, d):
        abc.ABCMeta.__init__(cls, name, b, d)
        if name == "OutputBase":
            return
        if not name.endswith("Output"):
            raise ValueError("Output classes must be named [Name]Output")
        config_name = name.rstrip("Output").lower()

        output_registry[config_name] = cls
        if hasattr(cls, "_aliases"):
            for alias in cls._aliases:
                #Do not over-ride superclass aliases
                if alias not in output_registry:
                    output_registry[alias] = cls

class CommentFileWrapper:
    """
    This little wrapper object is to turn an OutputBase object
    into a write-only file-like object where .write commands
    are turned into comments.

    This seemed cleaner than just adding a .write function to 
    OutputBase itself since it would look like that would write directly
    to file, not as comments.
    """
    def __init__(self, obj):
        self.obj=obj
    def write(self, text):
        self.obj.comment(text)

class OutputBase(metaclass=OutputMetaclass):
    def __init__(self):
        super(OutputBase,self).__init__()
        self._columns = []
        self.closed=False
        self.begun_sampling = False
        self.resumed = False

    @property
    def columns(self):
        return self._columns

    @columns.setter
    def columns(self, columns):
        self._columns = columns

    @columns.deleter
    def columns(self, columns):
        self._columns = []

    def add_column(self, name, dtype, comment=""):
        self.columns.append((name,dtype,comment))

    def del_column(self, column):
        if isinstance(column, int):
            self.columns.pop(column)
        elif isinstance(column, tuple):
            self.columns.remove(tuple)
        elif isinstance(column, str):
            self.columns.pop(self.column_index_for_name(column))
        else:
            raise TypeError("Unknown type of column to delete")

    def column_index_for_name(self, name):
        for i,(col_name,dtype,comment) in enumerate(self._columns):
            if name==col_name:
                return i

    @property
    def column_names(self):
        return [c[0] for c in self._columns]

    def comment(self, comment):
        """
        Save a comment.  Ordering will be preserved
        if you save multiple ones.
        """
        self._write_comment(comment.strip('\n'))

    def parameters(self, *param_groups):
        """ 
        Tell the outputter to save a vector of parameters.

        You can pass in either arrays/lists, which will
        be concatenated together, or mix in some scalars
        which will be included too.  Any number of arguments
        can be used.

        A typical pattern is for a sampler to send in three
        groups: sampled variable, extra outputs, and sampler 
        outputs (e.g. likelihoods or weights).

        The number of parameters and their types
        must match what the columns are.
        """
        #Check that the length is correct
        if self.closed:
            raise RuntimeError("Tried to write parameters to closed output")

        #This is very generic
        params = []
        for p in param_groups:
            if np.isscalar(p):
                params.append(p)
            else:
                params += list(p[:])
        if not len(params)==len(self._columns):
            raise ValueError("Sampler error - tried to save wrong number of parameters, or failed to set column names")

        #If this is our first sample then 
        if not self.begun_sampling:
            self._begun_sampling(params)
            self.begun_sampling=True
        #Pass to the subclasses to write output
        self._write_parameters(params)

    def reset_to_chain_start(self):
        """
        Seek the start of the chain so previous output can be overwritten.

        Subclasses can override this to actually do something.
        """
        pass

    def flush(self):
        """
        For supported output classes, flush all pending output
        """
        self._flush()

    def metadata(self, key, value, comment=""):
        """
        Save an item of metadata, with an optional comment.
        For many some output types this will only work before
        sampling has begun.
        """
        if self.closed:
            raise RuntimeError("Tried to write metadata info to closed output")
        self._write_metadata(key, value, comment)

    def final(self, key, value, comment=""):
        """
        Save an item of final metadata (like a convergence test).
        This is designed to be run after sampling has finished;
        doing otherwise might cause problems.
        """
        if self.closed:
            raise RuntimeError("Tried to write final info to closed output")
        self._write_final(key, value, comment)

    def name_for_sampler_resume_info(self):
        """
        Return a file name or directory name for an auxiliary file
        where the sampler can save information it needs to resume.

        This base class errors
        """
        raise NotImplementedError("You need to use a supported output format to use the resume feature with this sampler")


    def comment_file_wrapper(self):
        return CommentFileWrapper(self)

    @staticmethod
    def lock_file(f):
        fcntl.lockf(f, fcntl.LOCK_EX|fcntl.LOCK_NB)

    @staticmethod
    def unlock_file(f):
        fcntl.lockf(f, fcntl.LOCK_UN|fcntl.LOCK_NB)

    def close(self):
        self._close()
        self.closed=True

    def blinding_header(self):
        if self.resumed:
            return
        self.comment("")
        self.comment("Blank lines prevent accidental unblinding")
        for i in range(250):
            self.comment("")


    #These are the methods that subclasses should
    #implement.  _begun_sampling and _close are optional.
    #The others are mandatory

    def _begun_sampling(self, params):
        pass

    def _close(self):
        pass

    def _flush(self):
        pass

    @abc.abstractmethod
    def _write_parameters(self, params):
        pass

    @abc.abstractmethod
    def _write_metadata(self, key, value, comment):
        pass

    @abc.abstractmethod
    def _write_comment(self, comment):
        pass

    @abc.abstractmethod
    def _write_final(self, key, value, comment):
        pass

    @classmethod
    def from_options(cls, options, resume=False):
        """ This method should create an output object from the section of ini file it is given"""
        raise NotImplementedError("The format mode you tried to use is incomplete - sorry")

    @classmethod
    def load_from_options(cls, options):
        """ This method should load back data written using the format.
            It returns a tuple of (column_names, data, metadata, final_metadata)
            with column_names a list of strings
                 data: a list of numpy arrays (ordered by parameter, then sample)
                 metadata: a list of dictionaries
                 final_metadata: a list of dictionaries
            The latter three lists will have one element for any data chain
            identified within the data.
        """
        raise NotImplementedError("The format mode you tried to use is incomplete - sorry")



#coding: utf-8
from ..runtime import logs
import numpy as np



class Analytics(object):
    def __init__(self, params, pool=None):
        self.params = params
        self.pool = pool

        self.total_steps = 0
        nparam = len(params)
        self.means = np.zeros(nparam)
        self.m2 = np.zeros(nparam)
        self.cov_times_n = np.zeros((nparam,nparam))

    def add_traces(self, traces):
        if traces.shape[1] != len(self.params):
            raise RuntimeError("The number of traces added to Analytics "
                               "does not match the number of varied "
                               "parameters!")
        num = float(self.total_steps)
        for x in traces:
            num += 1.0
            delta = x - self.means
            old_means = self.means.copy()
            self.means += delta/num
            self.m2 += delta*(x - self.means)
            self.cov_times_n += np.outer(x-self.means, x-old_means)

        self.total_steps += traces.shape[0]

    def trace_means(self):
        if self.pool:
            return np.array(self.pool.gather(self.means)).T
        else:
            return self.means

    def trace_variances(self):
        if self.total_steps > 1:
            local_variance = self.m2 / float(self.total_steps-1)
            if self.pool:
                return np.array(self.pool.gather(local_variance)).T
            else:
                return local_variance
        return None

    def gelman_rubin(self):
        # takes current traces and returns
        if self.pool is None or not self.pool.size > 1:
            raise RuntimeError("Gelman-Rubin statistic is only "
                               "valid for multiple chains.")

        if self.total_steps == 0:
            raise RuntimeError("Gelman-Rubin statistic not "
                               "defined for 0-length chains.")

        # gather trace statistics to master process
        means = self.trace_means()
        variances = self.trace_variances()

        if self.pool.is_master():
            B_over_n = np.var(means, ddof=1, axis=1)
            B = B_over_n * self.total_steps
            W = np.mean(variances, axis=1)
            V = ((1. - 1./self.total_steps) * W +
                 (1. + 1./self.pool.size) * B_over_n)
            # TODO: check for 0-values in W
            Rhat = np.sqrt(V/W)
        else:
            Rhat = None

        Rhat = self.pool.bcast(Rhat)

        if self.pool.is_master():
            logs.important("Gelman-Rubin statistic:")
            for (p,R) in zip(self.params, Rhat):
                logs.important(f"    {p}   {R}")
            logs.important(f"Worst = {Rhat.max()}")

        return Rhat
